@@ -1,4 +1,4 @@
-#  Copyright 2023 Commonwealth Scientific and Industrial Research
+#  Copyright © 2018-2025 Commonwealth Scientific and Industrial Research
 #  Organisation (CSIRO) ABN 41 687 119 230.
 #
 #  Licensed under the Apache License, Version 2.0 (the "License");
@@ -14,6 +14,10 @@
 #  limitations under the License.
 
  #'@importFrom sparklyr j_invoke
+ #'@importFrom sparklyr spark_connection
+ #'@importFrom sparklyr j_invoke_new
+ #'@importFrom sparklyr invoke_static
+ #'@importFrom purrr map
 data_sources <- function(pc) {
   j_invoke(pc, "read")
 }
@@ -25,19 +29,13 @@ invoke_datasource <- function(pc, name, ...) {
       j_invoke(name, ...)
 }
 
-#' ImportMode
-#' 
-#' The following import modes are supported:
-#' \itemize{
-#'   \item{\code{OVERWRITE}: Overwrite any existing data.}
-#'   \item{\code{MERGE}: Merge the new data with the existing data based on resource ID.}
-#' }
-#'
-#' @export
-ImportMode <- list(
-    OVERWRITE = "overwrite",
-    MERGE = "merge"
-)
+to_java_list <- function(sc, elements) {
+  list <- j_invoke_new(sc, "java.util.ArrayList")
+  for (e in elements) {
+    j_invoke(list, "add", e)
+  }
+  list
+}
 
 #' SaveMode
 #'
@@ -47,12 +45,16 @@ ImportMode <- list(
 #'   \item{\code{APPEND}: Append the new data to the existing data.}
 #'   \item{\code{IGNORE}: Only save the data if the file does not already exist.}
 #'   \item{\code{ERROR}: Raise an error if the file already exists.}
+#'   \item{\code{MERGE}: Merge the new data with the existing data based on resource ID.}
 #' }
+#'
+#' @export
 SaveMode <- list(
     OVERWRITE = "overwrite",
     APPEND = "append",
     IGNORE = "ignore",
-    ERROR = "error"
+    ERROR = "error",
+    MERGE = "merge"
 )
 
 #' Create a data source from NDJSON
@@ -68,17 +70,16 @@ SaveMode <- list(
 #'   that it contains. Currently not implemented.
 #' @return A DataSource object that can be used to run queries against the data.
 #' 
-#' @seealso \href{https://pathling.csiro.au/docs/libraries/fhirpath-query#ndjson}{Pathling documentation - Reading NDJSON}
+#' @seealso \href{https://pathling.csiro.au/docs/libraries/io#ndjson}{Pathling documentation - Reading NDJSON}
 #'
 #' @export
 #' 
 #' @family data source functions
 #' 
-#' @examplesIf pathling_is_spark_installed()
-#' pc <- pathling_connect()
+#' @examples \dontrun{
 #' data_source <- pc %>% pathling_read_ndjson(pathling_examples('ndjson'))
 #' data_source %>% ds_read('Patient') %>% sparklyr::sdf_nrow()
-#' pathling_disconnect(pc)
+#' }
 pathling_read_ndjson <- function(pc, path, extension = "ndjson", file_name_mapper = NULL) {
   #See: issue #1601 (Implement file_name_mappers in R sparkly API)
   stopifnot(file_name_mapper == NULL)
@@ -95,18 +96,17 @@ pathling_read_ndjson <- function(pc, path, extension = "ndjson", file_name_mappe
 #' @param mime_type The MIME type of the bundles. Defaults to "application/fhir+json".
 #' @return A DataSource object that can be used to run queries against the data.
 #' 
-#' @seealso \href{https://pathling.csiro.au/docs/libraries/fhirpath-query#fhir-bundles}{Pathling documentation - Reading Bundles}
+#' @seealso \href{https://pathling.csiro.au/docs/libraries/io#fhir-bundles}{Pathling documentation - Reading Bundles}
 #' 
 #' @export
 #' 
 #' @family data source functions
 #' 
-#' @examplesIf pathling_is_spark_installed()
-#' pc <- pathling_connect()
+#' @examples \dontrun{
 #' data_source <- pc %>% pathling_read_bundles(pathling_examples('bundle-xml'),
 #'      c("Patient", "Observation"), MimeType$FHIR_XML)
 #' data_source %>% ds_read('Observation') %>% sparklyr::sdf_nrow()
-#' pathling_disconnect(pc)
+#' }
 pathling_read_bundles <- function(pc, path, resource_types, mime_type = MimeType$FHIR_JSON) {
 
   pc %>% invoke_datasource("bundles", as.character(path),
@@ -124,19 +124,18 @@ pathling_read_bundles <- function(pc, path, resource_types, mime_type = MimeType
 #'   and the values are the data frames containing the resource data.
 #' @return A DataSource object that can be used to run queries against the data.
 #' 
-#' @seealso \href{https://pathling.csiro.au/docs/libraries/fhirpath-query#datasets}{Pathling documentation - Reading datasets}
+#' @seealso \href{https://pathling.csiro.au/docs/libraries/io#datasets}{Pathling documentation - Reading datasets}
 #' 
 #' @export
 #' 
 #' @family data source functions
 #' 
-#' @examplesIf pathling_is_spark_installed()
-#' pc <- pathling_connect()
+#' @examples \dontrun{
 #' patient_df <- pc %>% pathling_example_resource('Patient')
 #' condition_df <- pc %>% pathling_example_resource('Condition')
 #' data_source <- pc %>% pathling_read_datasets(list(Patient = patient_df, Condition = condition_df))
 #' data_source %>% ds_read('Patient') %>% sparklyr::sdf_nrow()
-#' pathling_disconnect(pc)
+#' }
 pathling_read_datasets <- function(pc, resources) {
   resources <- as.list(resources)
   ds <- pc %>% invoke_datasource("datasets")
@@ -157,17 +156,16 @@ pathling_read_datasets <- function(pc, resources) {
 #' @param path The URI of the directory containing the Parquet tables.
 #' @return A DataSource object that can be used to run queries against the data.
 #' 
-#' @seealso \href{https://pathling.csiro.au/docs/libraries/fhirpath-query#parquet}{Pathling documentation - Reading Parquet}
+#' @seealso \href{https://pathling.csiro.au/docs/libraries/io#parquet}{Pathling documentation - Reading Parquet}
 #' 
 #' @export
 #' 
 #' @family data source functions
 #' 
-#' @examplesIf pathling_is_spark_installed()
-#' pc <- pathling_connect()
+#' @examples \dontrun{
 #' data_source <- pc %>% pathling_read_parquet(pathling_examples('parquet'))
 #' data_source %>% ds_read('Patient') %>% sparklyr::sdf_nrow()
-#' pathling_disconnect(pc)
+#' }
 pathling_read_parquet <- function(pc, path) {
   pc %>% invoke_datasource("parquet", as.character(path))
 }
@@ -182,17 +180,16 @@ pathling_read_parquet <- function(pc, path) {
 #' @param path The URI of the directory containing the Delta tables.
 #' @return A DataSource object that can be used to run queries against the data.
 #' 
-#' @seealso \href{https://pathling.csiro.au/docs/libraries/fhirpath-query#delta-lake}{Pathling documentation - Reading Delta}
+#' @seealso \href{https://pathling.csiro.au/docs/libraries/io#delta-lake}{Pathling documentation - Reading Delta}
 #' 
 #' @export
 #' 
 #' @family data source functions
 #' 
-#' @examplesIf pathling_is_spark_installed()
-#' pc <- pathling_connect()
+#' @examples \dontrun{
 #' data_source <- pc %>% pathling_read_delta(pathling_examples('delta'))
 #' data_source %>% ds_read('Patient') %>% sparklyr::sdf_nrow()
-#' pathling_disconnect(pc)
+#' }
 pathling_read_delta <- function(pc, path) {
   pc %>% invoke_datasource("delta", as.character(path))
 }
@@ -206,18 +203,16 @@ pathling_read_delta <- function(pc, path) {
 #' @param schema An optional schema name that should be used to qualify the table names.
 #' @return A DataSource object that can be used to run queries against the data.
 #' 
-#' @seealso \href{https://pathling.csiro.au/docs/libraries/fhirpath-query#managed-tables}{Pathling documentation - Reading managed tables}
+#' @seealso \href{https://pathling.csiro.au/docs/libraries/io#managed-tables}{Pathling documentation - Reading managed tables}
 #' 
 #' @export
 #'
 #' @family data source functions
 #' 
-#' @examplesIf pathling_is_spark_installed()
-#' pc <- pathling_connect()
-#' spark <- pathling_spark(pc)
+#' @examples \dontrun{
 #' data_source <- pc %>% pathling_read_tables()
 #' data_source %>% ds_read('Patient') %>% sparklyr::sdf_nrow()
-#' pathling_disconnect(pc)
+#' }
 pathling_read_tables <- function(pc, schema = NULL) {
   if (!is.null(schema)) {
     pc %>% invoke_datasource("tables", as.character(schema))
@@ -235,12 +230,11 @@ pathling_read_tables <- function(pc, schema = NULL) {
 #' 
 #' @export
 #'
-#' @examplesIf pathling_is_spark_installed()
-#' pc <- pathling_connect()
+#' @examples \dontrun{
 #' data_source <- pc %>% pathling_read_ndjson(pathling_examples('ndjson'))
 #' data_source %>% ds_read('Patient') %>% sparklyr::sdf_nrow()
 #' data_source %>% ds_read('Condition') %>% sparklyr::sdf_nrow()
-#' pathling_disconnect(pc)
+#' }
 ds_read <- function(ds, resource_code) {
   jdf <- j_invoke(ds, "read", resource_code)
   sdf_register(jdf)
@@ -251,10 +245,14 @@ data_sinks <- function(ds) {
 }
 
  #'@importFrom sparklyr invoke
-invoke_datasink <- function(ds, name, ...) {
-  ds %>%
-      data_sinks() %>%
-      invoke(name, ...)
+invoke_datasink <- function(ds, name, save_mode = NULL, ...) {
+  sink_builder <- ds %>% data_sinks()
+
+  if (!is.null(save_mode)) {
+    sink_builder <- sink_builder %>% invoke("saveMode", save_mode)
+  }
+
+  sink_builder %>% invoke(name, ...)
   return(invisible(NULL))
 }
 
@@ -272,16 +270,14 @@ invoke_datasink <- function(ds, name, ...) {
 #'
 #' @return No return value, called for side effects only.
 #' 
-#' @seealso \href{https://pathling.csiro.au/docs/libraries/fhirpath-query#ndjson-1}{Pathling documentation - Writing NDJSON}
+#' @seealso \href{https://pathling.csiro.au/docs/libraries/io#ndjson-1}{Pathling documentation - Writing NDJSON}
 #' 
-#' @examplesIf pathling_is_spark_installed()
-#' pc <- pathling_connect()
+#' @examples \dontrun{
 #' data_source <- pc %>% pathling_read_ndjson(pathling_examples('ndjson'))
 #' 
 #' # Write the data to a directory of NDJSON files.
 #' data_source %>% ds_write_ndjson(file.path(tempdir(), 'ndjson'))
-#' 
-#' pathling_disconnect(pc)
+#' }
 #' 
 #' @family data sink functions
 #' 
@@ -289,7 +285,7 @@ invoke_datasink <- function(ds, name, ...) {
 ds_write_ndjson <- function(ds, path, save_mode = SaveMode$ERROR, file_name_mapper = NULL) {
   #See: issue #1601 (Implement file_name_mappers in R sparkly API)
   stopifnot(file_name_mapper == NULL)
-  invoke_datasink(ds, "ndjson", path, save_mode)
+  invoke_datasink(ds, "ndjson", save_mode, path)
 }
 
 #' Write FHIR data to Parquet files
@@ -302,7 +298,7 @@ ds_write_ndjson <- function(ds, path, save_mode = SaveMode$ERROR, file_name_mapp
 #'
 #' @return No return value, called for side effects only.
 #' 
-#' @seealso \href{https://pathling.csiro.au/docs/libraries/fhirpath-query#parquet-1}{Pathling documentation - Writing Parquet}
+#' @seealso \href{https://pathling.csiro.au/docs/libraries/io#parquet-1}{Pathling documentation - Writing Parquet}
 #' 
 #' @examplesIf pathling_is_spark_installed()
 #' pc <- pathling_connect()
@@ -317,7 +313,7 @@ ds_write_ndjson <- function(ds, path, save_mode = SaveMode$ERROR, file_name_mapp
 #' 
 #' @export
 ds_write_parquet <- function(ds, path, save_mode = SaveMode$ERROR) {
-  invoke_datasink(ds, "parquet", path, save_mode)
+  invoke_datasink(ds, "parquet", save_mode, path)
 }
 
 #' Write FHIR data to Delta files
@@ -326,29 +322,29 @@ ds_write_parquet <- function(ds, path, save_mode = SaveMode$ERROR) {
 #'
 #' @param ds The DataSource object.
 #' @param path The URI of the directory to write the files to.
-#' @param import_mode The import mode to use when writing the data - "overwrite" will overwrite any 
+#' @param save_mode The save mode to use when writing the data - "overwrite" will overwrite any 
 #'      existing data, "merge" will merge the new data with the existing data based on resource ID.
 #'
 #' @return No return value, called for side effects only.
 #' 
-#' @seealso \href{https://pathling.csiro.au/docs/libraries/fhirpath-query#delta-lake-1}{Pathling documentation - Writing Delta}
+#' @seealso \href{https://pathling.csiro.au/docs/libraries/io#delta-lake-1}{Pathling documentation - Writing Delta}
 #' 
-#' @seealso \code{\link{ImportMode}}
+#' @seealso \code{\link{SaveMode}}
 #' 
 #' @examplesIf pathling_is_spark_installed()
 #' pc <- pathling_connect()
 #' data_source <- pc %>% pathling_read_ndjson(pathling_examples('ndjson'))
 #' 
 #' # Write the data to a directory of Delta files.
-#' data_source %>% ds_write_delta(file.path(tempdir(), 'delta'), import_mode = ImportMode$OVERWRITE)
+#' data_source %>% ds_write_delta(file.path(tempdir(), 'delta'), save_mode = SaveMode$OVERWRITE)
 #' 
 #' pathling_disconnect(pc)
 #' 
 #' @family data sink functions
 #' 
 #' @export
-ds_write_delta <- function(ds, path, import_mode = ImportMode$OVERWRITE) {
-  invoke_datasink(ds, "delta", path, import_mode)
+ds_write_delta <- function(ds, path, save_mode = SaveMode$OVERWRITE) {
+  invoke_datasink(ds, "delta", save_mode, path)
 }
 
 #' Write FHIR data to managed tables
@@ -357,14 +353,14 @@ ds_write_delta <- function(ds, path, import_mode = ImportMode$OVERWRITE) {
 #'
 #' @param ds The DataSource object.
 #' @param schema The name of the schema to write the tables to.
-#' @param import_mode The import mode to use when writing the data - "overwrite" will overwrite any 
+#' @param save_mode The save mode to use when writing the data - "overwrite" will overwrite any 
 #'      existing data, "merge" will merge the new data with the existing data based on resource ID.
 #' 
 #' @return No return value, called for side effects only.
 #' 
-#' @seealso \href{https://pathling.csiro.au/docs/libraries/fhirpath-query#managed-tables-1}{Pathling documentation - Writing managed tables}
+#' @seealso \href{https://pathling.csiro.au/docs/libraries/io#managed-tables-1}{Pathling documentation - Writing managed tables}
 #' 
-#' @seealso \code{\link{ImportMode}}
+#' @seealso \code{\link{SaveMode}}
 #'
 #' @examplesIf pathling_is_spark_installed()
 #' # Create a temporary warehouse location, which will be used when we call ds_write_tables().
@@ -382,7 +378,7 @@ ds_write_delta <- function(ds, path, import_mode = ImportMode$OVERWRITE) {
 #' data_source <- pc %>% pathling_read_ndjson(pathling_examples('ndjson'))
 #' 
 #' # Write the data to a set of Spark tables in the 'default' database.
-#' data_source %>% ds_write_tables("default", import_mode = ImportMode$MERGE)
+#' data_source %>% ds_write_tables("default", save_mode = SaveMode$MERGE)
 #' 
 #' pathling_disconnect(pc)
 #' unlink(temp_dir_path, recursive = TRUE)
@@ -390,10 +386,214 @@ ds_write_delta <- function(ds, path, import_mode = ImportMode$OVERWRITE) {
 #' @family data sink functions
 #' 
 #' @export
-ds_write_tables <- function(ds, schema = NULL, import_mode = ImportMode$OVERWRITE) {
+ds_write_tables <- function(ds, schema = NULL, save_mode = SaveMode$OVERWRITE) {
   if (!is.null(schema)) {
-    invoke_datasink(ds, "tables", import_mode, schema)
+    invoke_datasink(ds, "tables", save_mode, schema)
   } else {
-    invoke_datasink(ds, "tables", import_mode)
+    invoke_datasink(ds, "tables", save_mode)
   }
 }
+
+#' Create a data source from a FHIR Bulk Data Access API endpoint
+#'
+#' Creates a data source by downloading data from a FHIR server that implements the FHIR Bulk Data 
+#' Access API.
+#'
+#' @param pc The PathlingContext object.
+#' @param fhir_endpoint_url The URL of the FHIR server to export from.
+#' @param output_dir The directory to write the output files to.
+#' @param group_id Optional group ID for group-level export.
+#' @param patients Optional list of patient IDs for patient-level export.
+#' @param types List of FHIR resource types to include.
+#' @param output_format The format of the output data. Defaults to "application/fhir+ndjson".
+#' @param since Only include resources modified after this timestamp.
+#' @param elements List of FHIR elements to include.
+#' @param type_filters FHIR search queries to filter resources.
+#' @param include_associated_data Pre-defined set of FHIR resources to include.
+#' @param output_extension File extension for output files. Defaults to "ndjson".
+#' @param timeout Optional timeout duration in seconds.
+#' @param max_concurrent_downloads Maximum number of concurrent downloads. Defaults to 10.
+#' @param auth_config Optional authentication configuration list with the following possible elements:
+#'   \itemize{
+#'     \item{enabled: Whether authentication is enabled (default: FALSE)}
+#'     \item{client_id: The client ID to use for authentication}
+#'     \item{private_key_jwk: The private key in JWK format}
+#'     \item{client_secret: The client secret to use for authentication}
+#'     \item{token_endpoint: The token endpoint URL}
+#'     \item{use_smart: Whether to use SMART authentication (default: TRUE)}
+#'     \item{use_form_for_basic_auth: Whether to use form-based basic auth (default: FALSE)}
+#'     \item{scope: The scope to request}
+#'     \item{token_expiry_tolerance: The token expiry tolerance in seconds (default: 120)}
+#'   }
+#' @return A DataSource object that can be used to run queries against the data.
+#'
+#' @seealso \href{https://pathling.csiro.au/docs/libraries/io#fhir-bulk-data-api}{Pathling documentation - Reading from Bulk Data API}
+#'
+#' @export
+#'
+#' @family data source functions
+#'
+#' @examples \dontrun{
+#' pc <- pathling_connect()
+#' 
+#' # Basic system-level export
+#' data_source <- pc %>% pathling_read_bulk(
+#'   fhir_endpoint_url = "https://bulk-data.smarthealthit.org/fhir",
+#'   output_dir = "/tmp/bulk_export"
+#' )
+#' 
+#' # Group-level export with filters
+#' data_source <- pc %>% pathling_read_bulk(
+#'   fhir_endpoint_url = "https://bulk-data.smarthealthit.org/fhir", 
+#'   output_dir = "/tmp/bulk_export",
+#'   group_id = "group-1",
+#'   types = c("Patient", "Observation"),
+#'   elements = c("id", "status"),
+#'   since = as.POSIXct("2023-01-01")
+#' )
+#' 
+#' # Patient-level export with auth
+#' data_source <- pc %>% pathling_read_bulk(
+#'   fhir_endpoint_url = "https://bulk-data.smarthealthit.org/fhir",
+#'   output_dir = "/tmp/bulk_export", 
+#'   patients = c(
+#'     "123",  # Just the ID portion
+#'     "456"
+#'   ),
+#'   auth_config = list(
+#'     enabled = TRUE,
+#'     client_id = "my-client-id",
+#'     private_key_jwk = '{ "kty":"RSA", ...}',
+#'     scope = "system/*.read"
+#'   )
+#' )
+#' 
+#' pathling_disconnect(pc)
+#' }
+pathling_read_bulk <- function(pc,
+                               fhir_endpoint_url,
+                               output_dir,
+                               group_id = NULL,
+                               patients = NULL,
+                               types = NULL,
+                               output_format = "application/fhir+ndjson",
+                               since = NULL,
+                               elements = NULL,
+                               type_filters = NULL,
+                               include_associated_data = NULL,
+                               output_extension = "ndjson",
+                               timeout = NULL,
+                               max_concurrent_downloads = 10,
+                               auth_config = NULL) {
+  # Validate required parameters.
+  if (missing(fhir_endpoint_url)) {
+    stop("argument \"fhir_endpoint_url\" is missing")
+  }
+  if (missing(output_dir)) {
+    stop("argument \"output_dir\" is missing")
+  }
+
+  # Get the appropriate BulkExportClient builder based on export type
+  sc <- spark_connection(pc)
+  builder_class <- "au.csiro.fhir.export.BulkExportClient"
+
+  if (!is.null(group_id)) {
+    builder <- invoke_static(sc, builder_class, "groupBuilder", as.character(group_id))
+  } else if (!is.null(patients)) {
+    builder <- invoke_static(sc, builder_class, "patientBuilder")
+  } else {
+    builder <- invoke_static(sc, builder_class, "systemBuilder")
+  }
+
+  # Configure the basic parameters.
+  builder <- builder %>%
+      j_invoke("withFhirEndpointUrl", as.character(fhir_endpoint_url)) %>%
+      j_invoke("withOutputDir", as.character(output_dir)) %>%
+      j_invoke("withTypes", to_java_list(sc, types))
+
+  # Configure optional parameters if provided.
+  if (!is.null(output_format)) {
+    builder <- builder %>% j_invoke("withOutputFormat", as.character(output_format))
+  }
+  if (!is.null(since)) {
+    instant <- j_invoke_static(sc, "java.time.Instant", "ofEpochMilli", as.numeric(since) * 1000)
+    builder <- builder %>% j_invoke("withSince", instant)
+  }
+  if (!is.null(patients)) {
+    j_objects <- purrr::map(patients, function(r) j_invoke_static(sc, "au.csiro.fhir.model.Reference",
+                                                                  "of", r))
+    builder <- builder %>% j_invoke("withPatients", to_java_list(sc, j_objects))
+  }
+  if (!is.null(elements)) {
+    builder <- builder %>% j_invoke("withElements", to_java_list(sc, elements))
+  }
+  if (!is.null(type_filters)) {
+    builder <- builder %>% j_invoke("withTypeFilters", to_java_list(sc, type_filters))
+  }
+  if (!is.null(include_associated_data)) {
+    builder <- builder %>% j_invoke("withIncludeAssociatedData",
+                                    to_java_list(sc, include_associated_data))
+  }
+  if (!is.null(output_extension)) {
+    builder <- builder %>% j_invoke("withOutputExtension", as.character(output_extension))
+  }
+  if (!is.null(timeout)) {
+    j_object = j_invoke_static(sc, "java.time.Duration", "ofSeconds", as.numeric(timeout))
+    builder <- builder %>% j_invoke("withTimeout", j_object)
+  }
+  if (!is.null(max_concurrent_downloads)) {
+    builder <- builder %>% j_invoke("withMaxConcurrentDownloads", as.integer(max_concurrent_downloads))
+  }
+
+  # Configure authentication if provided.
+  if (!is.null(auth_config)) {
+    auth_builder <- j_invoke_new(sc, "au.csiro.pathling.auth.AuthConfig$AuthConfigBuilder")
+
+    # Set defaults to match Java class.
+    auth_builder <- auth_builder %>%
+        j_invoke("enabled", FALSE) %>%
+        j_invoke("useSMART", TRUE) %>%
+        j_invoke("useFormForBasicAuth", FALSE) %>%
+        j_invoke("tokenExpiryTolerance", 120L)
+
+    # Map R config to Java builder methods.
+    if (!is.null(auth_config$enabled))
+        auth_builder <- auth_builder %>% j_invoke("enabled", auth_config$enabled)
+
+    if (!is.null(auth_config$use_smart))
+        auth_builder <- auth_builder %>% j_invoke("useSMART", auth_config$use_smart)
+
+    if (!is.null(auth_config$token_endpoint))
+        auth_builder <- auth_builder %>% j_invoke("tokenEndpoint", auth_config$token_endpoint)
+
+    if (!is.null(auth_config$client_id))
+        auth_builder <- auth_builder %>% j_invoke("clientId", auth_config$client_id)
+
+    if (!is.null(auth_config$client_secret))
+        auth_builder <- auth_builder %>% j_invoke("clientSecret", auth_config$client_secret)
+
+    if (!is.null(auth_config$private_key_jwk))
+        auth_builder <- auth_builder %>% j_invoke("privateKeyJWK", auth_config$private_key_jwk)
+
+    if (!is.null(auth_config$use_form_for_basic_auth))
+        auth_builder <- auth_builder %>% j_invoke("useFormForBasicAuth",
+                                                  auth_config$use_form_for_basic_auth)
+
+    if (!is.null(auth_config$scope))
+        auth_builder <- auth_builder %>% j_invoke("scope", auth_config$scope)
+
+    if (!is.null(auth_config$token_expiry_tolerance))
+        auth_builder <- auth_builder %>% j_invoke("tokenExpiryTolerance",
+                                                  as.integer(auth_config$token_expiry_tolerance))
+
+    # Build auth config and add to builder.
+    auth_config_obj <- auth_builder %>% j_invoke("build")
+    builder <- builder %>% j_invoke("withAuthConfig", auth_config_obj)
+  }
+
+  # Build the BulkExportClient.
+  client <- builder %>% j_invoke("build")
+
+  # Pass the client to the bulk method.
+  pc %>% invoke_datasource("bulk", client)
+}
